/* gpModel.C
 * 
 * Copyright (C) 2009 Marcel Schumann
 * 
 * This file is part of QuEasy -- A Toolbox for Automated QSAR Model
 * Construction and Validation.
 * QuEasy is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * QuEasy is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

// -*- Mode: C++; tab-width: 2; -*-
// vi: set ts=2:
//
// 

#include <BALL/QSAR/gpModel.h>
#include <BALL/MATHS/LINALG/matrixInverter.h>

namespace BALL
{
	namespace QSAR
			{

		GPModel::GPModel(const QSARData& q, int k_type, double p1, double p2) : KernelModel(q, k_type, p1, p2) 
		{
			type_="GP";
			lambda_ = 0.001;
		}

		GPModel::GPModel(const QSARData& q, Vector<double>& w) : KernelModel(q, w) 
		{
			type_="GP";
			lambda_ = 0.001;
		}


		GPModel::GPModel(const QSARData& q, String s1, String s2) : KernelModel(q, s1, s2) 
		{
			type_="GP";
			lambda_ = 0.001;
		}

		GPModel::GPModel(const QSARData& q, const LinearModel& lm, int column) : KernelModel(q, lm, column) 
		{
			type_="GP";
			lambda_ = 0.001;
		}

		GPModel::~GPModel()
		{
		}


		void GPModel::train()
		{
			if (descriptor_matrix_.Ncols() == 0)
			{
				throw Exception::InconsistentUsage(__FILE__, __LINE__, "Data must be read into the model before training!"); 
			}
			kernel->calculateKernelMatrix(descriptor_matrix_, K_);
			
			Matrix<double> I(K_.Nrows(), K_.Nrows());
			I = 0;
			for (int i = 1; i < I.Nrows(); i++)
			{
				I(i, i) = 1;
			}
			
			MatrixInverter<double, StandardTraits> inverter(K_);
			inverter.computePseudoInverse();

			L_ = inverter.getPseudoInverse();  // dim: nxn

			training_result_ = L_*Y_; // B: one coefficient for each substance; dim: nxc
			
			calculateOffsets();
		}


		BALL::Vector<double> GPModel::predict(const vector<double> & substance, bool transform)
		{
			if (training_result_.Ncols() == 0)
			{
				throw Exception::InconsistentUsage(__FILE__, __LINE__, "Model must be trained before it can predict the activitiy of substances!"); 
			}
			input_ = getSubstanceVector(substance, transform); 
			
			kernel->calculateKernelVector(K_, input_, descriptor_matrix_, K_t_); // dim: 1xn
			
			Vector<double> res = K_t_*training_result_;
			//if (offsets_.getSize() == res.getSize()) res -= offsets_; 
			
			if (transform && y_transformations_.Ncols() != 0)
			{
				backTransformPrediction(res); 
			}
			return res;	
		}


		double GPModel::calculateStdErr()
		{
			Matrix<double> mx;
			Matrix<double> m1(1, input_.getSize());
			m1.copyVectorToRow(input_, 1);
			kernel->calculateKernelMatrix(m1, mx); // k(x*, x*), dim: 1x1
			
			double sum = 0;
			for (uint i = 1; i <= K_t_.getSize(); i++)
			{
				for (uint j = 1; j <= K_t_.getSize(); j++)
				{
					sum += K_t_(i)*K_t_(j)*L_(i, j); // k(x*, _i)*k(x*, x_j)*L_ij
				}
			}
			double res = sqrt(abs(mx(1, 1)-sum));
			return res;	
		}


		void GPModel::setParameters(vector<double>& v)
		{
			if (v.size() != 1)
			{
				String c = "Wrong number of model parameters! Needed: 1;";
				c = c+" given: "+String(v.size());
				throw Exception::ModelParameterError(__FILE__, __LINE__, c.c_str());
			}
			lambda_ = v[0];
		}

		vector<double> GPModel::getParameters() const
		{
			vector<double> d;
			d.push_back(lambda_);
			return d;
		}
	}
}
